@@ -11,13 +11,7 @@ from torch.nn.modules import Module
 import torch.multiprocessing as mp
 import torch.distributed as dist
 from torch.nn.parallel import DistributedDataParallel
-try:
-    from apex import amp
-    from apex.parallel import DistributedDataParallel as ApexDistributedDataParallel
-    from aitoolbox.torchtrain.parallel import TTApexDistributedDataParallel
-    APEX_AVAILABLE = True
-except ImportError:
-    APEX_AVAILABLE = False
+import torch.cuda.amp as amp
 try:
     import deepspeed
     from aitoolbox.torchtrain.parallel import TTDeepSpeedLight
@@ -33,6 +27,7 @@ from aitoolbox.torchtrain.data.batch_model_feed_defs import AbstractModelFeedDef
 from aitoolbox.torchtrain.tl_components.callback_handler import CallbacksHandler
 from aitoolbox.torchtrain.tl_components.ddp_handler import DDPHandler
 from aitoolbox.torchtrain.callbacks.model_save import ModelCheckpoint, ModelTrainEndSave
+from aitoolbox.torchtrain.schedulers.basic import AbstractScheduler
 from aitoolbox.experiment.training_history import TrainingHistory
 from aitoolbox.torchtrain.tl_components.model_prediction_store import ModelPredictionStore
 from aitoolbox.torchtrain.tl_components.message_passing import MessageService
@@ -45,7 +40,7 @@ class TrainLoop:
                  train_loader, validation_loader, test_loader,
                  optimizer, criterion,
                  collate_batch_pred_fn=append_predictions, pred_transform_fn=torch_cat_transf,
-                 end_auto_eval=True,
+                 end_auto_eval=True, lazy_experiment_save=True,
                  gpu_mode='single', cuda_device_idx=None, use_amp=False):
         """Core PyTorch TrainLoop supporting the model training and target prediction
 
@@ -59,7 +54,7 @@ class TrainLoop:
             validation_loader (torch.utils.data.DataLoader or None): data loader for validation data set
             test_loader (torch.utils.data.DataLoader or None): data loader for test data set
             optimizer (torch.optim.optimizer.Optimizer or MultiOptimizer): optimizer algorithm.
-            criterion (torch.nn.modules.loss._Loss or MultiLoss): criterion criterion during the training procedure.
+            criterion (torch.nn.modules.loss._Loss or MultiLoss or None): criterion during the training procedure
             collate_batch_pred_fn (callable): collate function transforming batch predictions as they come out from the
                 model
             pred_transform_fn (callable): function transforming all the produced predictions after all the batches have
@@ -68,6 +63,9 @@ class TrainLoop:
                 loss calculations. This is useful when conducting very costly experiments to save on compute time.
                 Specify either True/False boolean to always run or never run after each epoch or specify an int to
                 execute only every specified number of epochs.
+            lazy_experiment_save (bool): when in lazy mode experiment tracking components will create the experiment
+                folder only after some training results are available (possibly at the end of the first epoch) instead
+                of at the beginning of training.
             gpu_mode (str): GPU training mode selection. TrainLoop supports different GPU training modes by
                 specifying one of the following:
 
@@ -77,14 +75,12 @@ class TrainLoop:
                 * ``'deepspeed'``: training via the Microsoft DeepSpeed
 
             cuda_device_idx (int or None): CUDA device index used when training on multiple GPUs
-            use_amp (bool or dict): use Nvidia Apex 16-bit Automatic Mixed Precision (AMP)
+            use_amp (bool or dict): use 16-bit Automatic Mixed Precision (AMP)
 
                 To switch to AMP mode either:
 
-                * set this parameter to ``True`` to use default AMP initialization parameters
-                * provide custom Apex AMP initialization parameters as a dict as this parameter
-
-                Available AMP initialization parameters: https://nvidia.github.io/apex/amp.html#apex.amp.initialize
+                * set this parameter to ``True`` to use default AMP ``torch.cuda.amp.GradScaler`` initialization params
+                * provide custom AMP ``torch.cuda.amp.GradScaler`` initialization parameters as a dict as this parameter
         """
         if isinstance(model, TTModel) or isinstance(model, TTDataParallel):
             self.model = model
@@ -104,12 +100,14 @@ class TrainLoop:
         self.collate_batch_pred_fn = collate_batch_pred_fn
         self.pred_transform_fn = pred_transform_fn
         self.end_auto_eval = end_auto_eval
+        self.lazy_experiment_save = lazy_experiment_save
 
         self.num_optimizers = 1 if not isinstance(self.optimizer, MultiOptimizer) else len(self.optimizer)
 
         self.gpu_mode = gpu_mode
         self.use_amp = use_amp is True or type(use_amp) == dict
-        self.amp_params = {} if use_amp is True else use_amp
+        self.amp_scaler_init = {} if use_amp is True else use_amp
+        self.amp_scaler = amp.GradScaler(**self.amp_scaler_init) if self.use_amp and self.gpu_mode != 'ddp' else None
         self.use_deepspeed = False
 
         USE_CUDA = torch.cuda.is_available()
@@ -148,9 +146,6 @@ class TrainLoop:
         if self.gpu_mode not in ['single', 'dp', 'ddp', 'deepspeed']:
             raise ValueError("gpu_mode parameter set to the non-supported value. Can use only the following values: "
                              "'single', 'dp', 'ddp' and 'deepspeed'")
-        if self.use_amp and not APEX_AVAILABLE:
-            raise ValueError('Trying to use Nvidia Apex AMP for 16-bit mixed precision. However, Nvidia Apex is not'
-                             'installed.')
 
     def fit(self, num_epochs, callbacks=None, grad_accumulation=1, **kwargs):
         """Train the model using the train loop
@@ -206,10 +201,8 @@ class TrainLoop:
         self.callbacks_handler.register_callbacks(callbacks)
 
         self.model = self.model.to(self.device)
-        self.criterion = self.criterion.to(self.device)
-        # Initialize AMP when training with Nvidia APEX mixed precision
-        if self.use_amp and not self.ddp_training_mode:
-            self.model, self.optimizer = amp.initialize(self.model, self.optimizer, **self.amp_params)
+        if self.criterion is not None:
+            self.criterion = self.criterion.to(self.device)
 
         self.model.train()
 
@@ -231,7 +224,7 @@ class TrainLoop:
                 # Iterate over potentially multiple optimizers
                 for optimizer_idx in range(self.num_optimizers):
                     # Backward pass through the model
-                    self._backward_pass(loss_batch, optimizer_idx)
+                    self._backward_pass(loss_batch, iteration, optimizer_idx)
                     if self.grad_cb_used:
                         self.callbacks_handler.execute_gradient_update(optimizer_idx)
 
@@ -240,6 +233,9 @@ class TrainLoop:
                     # Optimizer zero grad
                     self._optimizer_zero_grad(iteration, grad_accumulation, optimizer_idx)
 
+                if self.use_amp:
+                    self.amp_scaler.update()
+
                 self.callbacks_handler.execute_batch_end()
 
             # Automatic end of epoch code - reports the train and if available validation loss and executes callbacks
@@ -271,44 +267,50 @@ class TrainLoop:
         Returns:
             loss: loss calculated on current batch
         """
-        if self.batch_model_feed_def is None:
-            loss_batch = self.model.get_loss(batch_data, self.criterion, self.device)
-        else:
-            loss_batch = self.batch_model_feed_def.get_loss(self.model, batch_data, self.criterion, self.device)
+        with amp.autocast(enabled=self.use_amp):
+            if self.batch_model_feed_def is None:
+                loss_batch = self.model.get_loss(batch_data, self.criterion, self.device)
+            else:
+                loss_batch = self.batch_model_feed_def.get_loss(self.model, batch_data,
+                                                                self.criterion, self.device)
+            loss_batch_log = loss_batch
 
-        self.loss_batch_accum.append(loss_batch.item())
+            # Need to divide by the number of accumulation steps if our loss is averaged over the training samples
+            loss_batch = loss_batch / grad_accumulation
 
-        # Need to divide by the number of accumulation steps if our loss is averaged over the training samples
-        loss_batch = loss_batch / grad_accumulation
+        self.loss_batch_accum.append(loss_batch_log.item())
 
         return loss_batch
 
-    def _backward_pass(self, loss_batch, optimizer_idx):
+    def _backward_pass(self, loss_batch, iteration, optimizer_idx):
         """Execute backward pass from the current batch loss
 
         Args:
             loss_batch: loss calculated on current batch
+            iteration (int): current iteration index
             optimizer_idx (int): index of the current optimizer. Mostly useful when using multiple optimizers. When
                 only a single optimizer is used this parameter can be ignored.
 
         Returns:
             None
         """
-        if self.use_amp:
-            if not isinstance(loss_batch, MultiLoss):
-                # Single loss Apex AMP calculation
-                with amp.scale_loss(loss_batch, self.optimizer) as scaled_loss:
-                    scaled_loss.backward()
-            else:
-                # Multi-loss Apex AMP calculation
-                loss_batch.backward_amp(self.optimizer, optimizer_idx)
-        elif self.use_deepspeed:
+        if self.use_deepspeed:
             self.model.backward(loss_batch)
         else:
             if not isinstance(loss_batch, MultiLoss):
+                # if not self.use_amp:
+                #     loss_batch.backward()
+                # else:
+                #     self.amp_scaler.scale(loss_batch).backward()
+
+                # TODO: Make sure this is equivalent to the commented out code block above
+                if self.use_amp and self.amp_scaler is not None:
+                    loss_batch = self.amp_scaler.scale(loss_batch)
+
                 loss_batch.backward()
             else:
-                loss_batch.backward(optimizer_idx)
+                # Non-AMP or AMP backward are done under the hood in the MultiLoss wrap
+                loss_batch.backward(optimizer_idx, iteration, self.amp_scaler)
 
     def _optimizer_step(self, iteration, grad_accumulation, optimizer_idx):
         """Execute the optimizer step
@@ -328,9 +330,13 @@ class TrainLoop:
                 self.model.step()
             else:
                 if not isinstance(self.optimizer, MultiOptimizer):
-                    self.optimizer.step()
+                    if not self.use_amp:
+                        self.optimizer.step()
+                    else:
+                        self.amp_scaler.step(self.optimizer)
                 else:
-                    self.optimizer.step(optimizer_idx)
+                    # Non-AMP or AMP optimizer step are done under the hood in the MultiOptimizer wrap
+                    self.optimizer.step(optimizer_idx, iteration, self.amp_scaler)
 
             if self.grad_cb_used:
                 self.callbacks_handler.execute_optimizer_step()
@@ -353,7 +359,7 @@ class TrainLoop:
                 if not isinstance(self.optimizer, MultiOptimizer):
                     self.optimizer.zero_grad()
                 else:
-                    self.optimizer.zero_grad(optimizer_idx)
+                    self.optimizer.zero_grad(optimizer_idx, iteration)
 
     def auto_execute_end_of_epoch(self):
         """Basic performance evaluation executed by default at the end of each epoch
@@ -363,7 +369,7 @@ class TrainLoop:
         Returns:
             None
         """
-        loss_parsed = self._parse_loss(self.loss_batch_accum)
+        loss_parsed = self.parse_loss(self.loss_batch_accum)
         self._print_save_loss(loss_parsed,
                               loss_type_name='accumulated_loss',
                               loss_print_description='AVG BATCH ACCUMULATED TRAIN LOSS')
@@ -391,7 +397,7 @@ class TrainLoop:
             # has been turned off
             self._print_save_loss(test_loss, loss_type_name='train_end_test_loss', loss_print_description='TEST LOSS')
 
-    def _parse_loss(self, loss_record):
+    def parse_loss(self, loss_record):
         """Helper function to process different possible loss formats
 
         Primarily useful for parsing between single loss representation and the multi-loss representation.
@@ -517,22 +523,24 @@ class TrainLoop:
             float or dict: loss, in the case of multi loss, the dict gets returned
         """
         self.model = self.model.to(self.device)
-        self.criterion = self.criterion.to(self.device)
+        if self.criterion is not None:
+            self.criterion = self.criterion.to(self.device)
 
         self.model.eval()
         loss_avg = []
 
         with torch.no_grad():
             for batch_data in tqdm(data_loader):
-                if self.batch_model_feed_def is None:
-                    loss_batch = self.model.get_loss_eval(batch_data, self.criterion, self.device)
-                else:
-                    loss_batch = self.batch_model_feed_def.get_loss_eval(self.model, batch_data, self.criterion,
-                                                                         self.device)
+                with amp.autocast(enabled=self.use_amp):
+                    if self.batch_model_feed_def is None:
+                        loss_batch = self.model.get_loss_eval(batch_data, self.criterion, self.device)
+                    else:
+                        loss_batch = self.batch_model_feed_def.get_loss_eval(self.model, batch_data, self.criterion,
+                                                                             self.device)
 
                 loss_avg.append(loss_batch.item())
 
-            loss_avg = self._parse_loss(loss_avg)
+            loss_avg = self.parse_loss(loss_avg)
 
         self.model.train()
 
@@ -609,11 +617,12 @@ class TrainLoop:
 
         with torch.no_grad():
             for batch_data in tqdm(data_loader):
-                if self.batch_model_feed_def is None:
-                    y_pred_batch, y_test_batch, metadata_batch = self.model.get_predictions(batch_data, self.device)
-                else:
-                    y_pred_batch, y_test_batch, metadata_batch = \
-                        self.batch_model_feed_def.get_predictions(self.model, batch_data, self.device)
+                with amp.autocast(enabled=self.use_amp):
+                    if self.batch_model_feed_def is None:
+                        y_pred_batch, y_test_batch, metadata_batch = self.model.get_predictions(batch_data, self.device)
+                    else:
+                        y_pred_batch, y_test_batch, metadata_batch = \
+                            self.batch_model_feed_def.get_predictions(self.model, batch_data, self.device)
 
                 y_pred = self.collate_batch_pred_fn(y_pred_batch, y_pred)
                 y_test = self.collate_batch_pred_fn(y_test_batch, y_test)
@@ -646,6 +655,16 @@ class TrainLoop:
         """
         self.train_history.insert_single_result_into_history(metric_name, metric_result)
 
+    def get_schedulers(self):
+        """Get the registered schedulers
+
+        Schedulers in TrainLoop training are implemented as callbacks under the hood.
+
+        Returns:
+            list: list of scheduler (callbacks)
+        """
+        return [cb for cb in self.callbacks if isinstance(cb, AbstractScheduler)]
+
     def _train_dp(self, num_epochs, callbacks=None, grad_accumulation=1, dp_model_args=None):
         """Train the model on multi-GPU with DataParallel auto wrapping
 
@@ -655,9 +674,6 @@ class TrainLoop:
             grad_accumulation (int): number of batches the gradients are accumulated before updating weights
             dp_model_args (dict or None): parameters for :class:`aitoolbox.torchtrain.parallel.TTDataParallel` /
                 ``nn.DataParallel`` DP model wrap.
-                Probably the most common optional parameter to set is ``TTDataParallel``'s ``add_model_attributes``
-                list. In this list the user can list any additional TTModel attributes which need to be transferred to
-                the TTDataParallel level to enable their use in the transferred/exposed class methods.
 
         Returns:
             TTDataParallel or nn.DataParallel: trained model
@@ -683,11 +699,9 @@ class TrainLoop:
             num_epochs (int): how many epochs the network will be trained
             callbacks (list or None): callbacks that are executed during the training run
             grad_accumulation (int): number of batches the gradients are accumulated before updating weights
-            ddp_model_args (dict or None): parameters for DistributedDataParallel / APEX DistributedDataParallel model
+            ddp_model_args (dict or None): parameters for DistributedDataParallel model
                 Available parameters for DistributedDataParallel:
                     https://pytorch.org/docs/master/nn.html#torch.nn.parallel.DistributedDataParallel
-                Available parameters for APEX DistributedDataParallel:
-                    https://nvidia.github.io/apex/parallel.html#apex.parallel.DistributedDataParallel
             in_process_data_load (AbstractCallback or list or None):
                 in-process data loading logic implemented as a torchtrain callback. The logic should be placed inside
                 the on_multiprocess_start() callback function.
@@ -751,30 +765,23 @@ class TrainLoop:
 
         # Move to the GPU belonging to the process
         self.model = self.model.to(self.device)
-        self.criterion = self.criterion.to(self.device)
+        if self.criterion is not None:
+            self.criterion = self.criterion.to(self.device)
 
-        # Optionally initialize APEX
-        # Not using AMP initialization at the start of the fit() fn because in the multi-GPU setting the model has to
-        # be first AMP-initialized before it's wrapped into (APEX) DistributedDataParallel.
+        # Optionally initialize AMP scaler inside each of the processes
         if self.use_amp:
-            self.model, self.optimizer = amp.initialize(self.model, self.optimizer, **self.amp_params)
+            self.amp_scaler = amp.GradScaler(**self.amp_scaler_init)
 
         # Wrap models into DDP module
         if isinstance(self.model, TTModel):
-            if not self.use_amp:
-                self.model = TTDistributedDataParallel(self.model, device_ids=[gpu], **ddp_args['ddp_model_args'])
-            else:
-                self.model = TTApexDistributedDataParallel(self.model, **ddp_args['ddp_model_args'])
+            self.model = TTDistributedDataParallel(self.model, device_ids=[gpu], **ddp_args['ddp_model_args'])
         else:
-            if not self.use_amp:
-                self.model = DistributedDataParallel(self.model, device_ids=[gpu], **ddp_args['ddp_model_args'])
-            else:
-                self.model = ApexDistributedDataParallel(self.model, **ddp_args['ddp_model_args'])
+            self.model = DistributedDataParallel(self.model, device_ids=[gpu], **ddp_args['ddp_model_args'])
 
         self._train(num_epochs, callbacks, grad_accumulation)
 
     def _train_deepspeed(self, deepspeed_args, num_epochs, callbacks=None,
-                         add_model_attributes=None, **ds_model_args):
+                         **ds_model_args):
         """Train the model using Microsoft DeepSpeed package
 
         Before starting the training the DeepSpeed library needs to be installed on the machine. Find the installation
@@ -788,8 +795,6 @@ class TrainLoop:
                 A dictionary containing local_rank and deepspeed_config file location.
             num_epochs (int): how many epochs the network will be trained
             callbacks (list): callbacks that are executed during the training run
-            add_model_attributes (list or tuple or None): additional TTModel attributes which need to be transferred to
-                the TTDataParallel level to enable their use in the transferred/exposed class methods
             **ds_model_args: additional parameters for the underlying ``deepspeed.DeepSpeedLight`` class
 
                 Possible arguments: https://deepspeed.readthedocs.io/en/latest/initialize.html
@@ -807,7 +812,7 @@ class TrainLoop:
 
         self.model = TTDeepSpeedLight(
             args=deepspeed_args,
-            model=self.model, model_parameters=self.model.parameters(), add_model_attributes=add_model_attributes,
+            model=self.model, model_parameters=self.model.parameters(),
             training_data=self.train_loader.dataset,
             **ds_model_args
         )
@@ -842,7 +847,7 @@ class TrainLoopCheckpoint(TrainLoop):
                  cloud_save_mode='s3', bucket_name='model-result', cloud_dir_prefix='', source_dirs=(),
                  rm_subopt_local_models=False, num_best_checkpoints_kept=2,
                  collate_batch_pred_fn=append_predictions, pred_transform_fn=torch_cat_transf,
-                 end_auto_eval=True,
+                 end_auto_eval=True, lazy_experiment_save=True,
                  gpu_mode='single', cuda_device_idx=None, use_amp=False):
         """TrainLoop with the automatic model check-pointing at the end of each epoch
 
@@ -852,7 +857,7 @@ class TrainLoopCheckpoint(TrainLoop):
             validation_loader (torch.utils.data.DataLoader or None): data loader for validation data set
             test_loader (torch.utils.data.DataLoader or None): data loader for test data set
             optimizer (torch.optim.optimizer.Optimizer or MultiOptimizer): optimizer algorithm.
-            criterion (torch.nn.modules.loss._Loss or MultiLoss): criterion criterion during the training procedure.
+            criterion (torch.nn.modules.loss._Loss or MultiLoss or None): criterion during the training procedure
             project_name (str): root name of the project
             experiment_name (str): name of the particular experiment
             local_model_result_folder_path (str): root local path where project folder will be created
@@ -880,6 +885,9 @@ class TrainLoopCheckpoint(TrainLoop):
                 loss calculations. This is useful when conducting very costly experiments to save on compute time.
                 Specify either True/False boolean to always run or never run after each epoch or specify an int to
                 execute only every specified number of epochs.
+            lazy_experiment_save (bool): when in lazy mode experiment tracking components will create the experiment
+                folder only after some training results are available (possibly at the end of the first epoch) instead
+                of at the beginning of training.
             gpu_mode (str): GPU training mode selection. TrainLoop supports different GPU training modes by
                 specifying one of the following:
 
@@ -889,18 +897,17 @@ class TrainLoopCheckpoint(TrainLoop):
                 * ``'deepspeed'``: training via the Microsoft DeepSpeed
 
             cuda_device_idx (int or None): CUDA device index used when training on multiple GPUs
-            use_amp (bool or dict): use Nvidia Apex 16-bit Automatic Mixed Precision (AMP)
+            use_amp (bool or dict): use 16-bit Automatic Mixed Precision (AMP)
 
                 To switch to AMP mode either:
 
-                * set this parameter to ``True`` to use default AMP initialization parameters
-                * provide custom Apex AMP initialization parameters as a dict as this parameter
-
-                Available AMP initialization parameters: https://nvidia.github.io/apex/amp.html#apex.amp.initialize
+                * set this parameter to ``True`` to use default AMP ``torch.cuda.amp.GradScaler`` initialization params
+                * provide custom AMP ``torch.cuda.amp.GradScaler`` initialization parameters as a dict as this parameter
         """
         TrainLoop.__init__(self, model, train_loader, validation_loader, test_loader, optimizer, criterion,
                            collate_batch_pred_fn, pred_transform_fn,
-                           end_auto_eval, gpu_mode, cuda_device_idx, use_amp)
+                           end_auto_eval, lazy_experiment_save,
+                           gpu_mode, cuda_device_idx, use_amp)
         self.project_name = project_name
         self.experiment_name = experiment_name
         self.local_model_result_folder_path = os.path.expanduser(local_model_result_folder_path)
@@ -933,7 +940,7 @@ class TrainLoopEndSave(TrainLoop):
                  hyperparams, val_result_package=None, test_result_package=None,
                  cloud_save_mode='s3', bucket_name='model-result', cloud_dir_prefix='', source_dirs=(),
                  collate_batch_pred_fn=append_predictions, pred_transform_fn=torch_cat_transf,
-                 end_auto_eval=True,
+                 end_auto_eval=True, lazy_experiment_save=True,
                  gpu_mode='single', cuda_device_idx=None, use_amp=False):
         """TrainLoop with the model performance evaluation and final model saving at the end of the training process
 
@@ -943,7 +950,7 @@ class TrainLoopEndSave(TrainLoop):
             validation_loader (torch.utils.data.DataLoader or None): data loader for validation data set
             test_loader (torch.utils.data.DataLoader or None): data loader for test data set
             optimizer (torch.optim.optimizer.Optimizer or MultiOptimizer): optimizer algorithm.
-            criterion (torch.nn.modules.loss._Loss or MultiLoss): criterion criterion during the training procedure.
+            criterion (torch.nn.modules.loss._Loss or MultiLoss or None): criterion during the training procedure
             project_name (str): root name of the project
             experiment_name (str): name of the particular experiment
             local_model_result_folder_path (str): root local path where project folder will be created
@@ -970,6 +977,9 @@ class TrainLoopEndSave(TrainLoop):
                 loss calculations. This is useful when conducting very costly experiments to save on compute time.
                 Specify either True/False boolean to always run or never run after each epoch or specify an int to
                 execute only every specified number of epochs.
+            lazy_experiment_save (bool): when in lazy mode experiment tracking components will create the experiment
+                folder only after some training results are available (possibly at the end of the first epoch) instead
+                of at the beginning of training.
             gpu_mode (str): GPU training mode selection. TrainLoop supports different GPU training modes by
                 specifying one of the following:
 
@@ -979,18 +989,17 @@ class TrainLoopEndSave(TrainLoop):
                 * ``'deepspeed'``: training via the Microsoft DeepSpeed
 
             cuda_device_idx (int or None): CUDA device index used when training on multiple GPUs
-            use_amp (bool or dict): use Nvidia Apex 16-bit Automatic Mixed Precision (AMP)
+            use_amp (bool or dict): use 16-bit Automatic Mixed Precision (AMP)
 
                 To switch to AMP mode either:
 
-                * set this parameter to ``True`` to use default AMP initialization parameters
-                * provide custom Apex AMP initialization parameters as a dict as this parameter
-
-                Available AMP initialization parameters: https://nvidia.github.io/apex/amp.html#apex.amp.initialize
+                * set this parameter to ``True`` to use default AMP ``torch.cuda.amp.GradScaler`` initialization params
+                * provide custom AMP ``torch.cuda.amp.GradScaler`` initialization parameters as a dict as this parameter
         """
         TrainLoop.__init__(self, model, train_loader, validation_loader, test_loader, optimizer, criterion,
                            collate_batch_pred_fn, pred_transform_fn,
-                           end_auto_eval, gpu_mode, cuda_device_idx, use_amp)
+                           end_auto_eval, lazy_experiment_save,
+                           gpu_mode, cuda_device_idx, use_amp)
         self.project_name = project_name
         self.experiment_name = experiment_name
         self.local_model_result_folder_path = os.path.expanduser(local_model_result_folder_path)
@@ -1043,7 +1052,7 @@ class TrainLoopCheckpointEndSave(TrainLoopEndSave):
                  cloud_save_mode='s3', bucket_name='model-result', cloud_dir_prefix='', source_dirs=(),
                  rm_subopt_local_models=False, num_best_checkpoints_kept=2,
                  collate_batch_pred_fn=append_predictions, pred_transform_fn=torch_cat_transf,
-                 end_auto_eval=True,
+                 end_auto_eval=True, lazy_experiment_save=True,
                  gpu_mode='single', cuda_device_idx=None, use_amp=False):
         """TrainLoop both saving model check-pointing at the end of each epoch and model performance reporting
             and model saving at the end of the training process
@@ -1054,7 +1063,7 @@ class TrainLoopCheckpointEndSave(TrainLoopEndSave):
             validation_loader (torch.utils.data.DataLoader or None): data loader for validation data set
             test_loader (torch.utils.data.DataLoader or None): data loader for test data set
             optimizer (torch.optim.optimizer.Optimizer or MultiOptimizer): optimizer algorithm.
-            criterion (torch.nn.modules.loss._Loss or MultiLoss): criterion criterion during the training procedure.
+            criterion (torch.nn.modules.loss._Loss or MultiLoss or None): criterion during the training procedure
             project_name (str): root name of the project
             experiment_name (str): name of the particular experiment
             local_model_result_folder_path (str): root local path where project folder will be created
@@ -1086,6 +1095,9 @@ class TrainLoopCheckpointEndSave(TrainLoopEndSave):
                 loss calculations. This is useful when conducting very costly experiments to save on compute time.
                 Specify either True/False boolean to always run or never run after each epoch or specify an int to
                 execute only every specified number of epochs.
+            lazy_experiment_save (bool): when in lazy mode experiment tracking components will create the experiment
+                folder only after some training results are available (possibly at the end of the first epoch) instead
+                of at the beginning of training.
             gpu_mode (str): GPU training mode selection. TrainLoop supports different GPU training modes by
                 specifying one of the following:
 
@@ -1095,14 +1107,12 @@ class TrainLoopCheckpointEndSave(TrainLoopEndSave):
                 * ``'deepspeed'``: training via the Microsoft DeepSpeed
 
             cuda_device_idx (int or None): CUDA device index used when training on multiple GPUs
-            use_amp (bool or dict): use Nvidia Apex 16-bit Automatic Mixed Precision (AMP)
+            use_amp (bool or dict): use 16-bit Automatic Mixed Precision (AMP)
 
                 To switch to AMP mode either:
 
-                * set this parameter to ``True`` to use default AMP initialization parameters
-                * provide custom Apex AMP initialization parameters as a dict as this parameter
-
-                Available AMP initialization parameters: https://nvidia.github.io/apex/amp.html#apex.amp.initialize
+                * set this parameter to ``True`` to use default AMP ``torch.cuda.amp.GradScaler`` initialization params
+                * provide custom AMP ``torch.cuda.amp.GradScaler`` initialization parameters as a dict as this parameter
         """
         if 'experiment_file_path' not in hyperparams:
             hyperparams['experiment_file_path'] = inspect.getframeinfo(inspect.currentframe().f_back).filename
@@ -1115,7 +1125,8 @@ class TrainLoopCheckpointEndSave(TrainLoopEndSave):
                                   hyperparams, val_result_package, test_result_package,
                                   cloud_save_mode, bucket_name, cloud_dir_prefix, source_dirs,
                                   collate_batch_pred_fn, pred_transform_fn,
-                                  end_auto_eval, gpu_mode, cuda_device_idx, use_amp)
+                                  end_auto_eval, lazy_experiment_save,
+                                  gpu_mode, cuda_device_idx, use_amp)
         self.rm_subopt_local_models = rm_subopt_local_models
 
         self.callbacks_handler.register_callbacks([
